@@ -95,6 +95,22 @@ if sodium_init() < 0:
     raise RuntimeError('sodium_init() call failed!')
 
 # Define constants
+try:
+    crypto_box_SEALBYTES = nacl.crypto_box_sealbytes()
+    HAS_SEAL = True
+except AttributeError:
+    HAS_SEAL = False
+try:
+    crypto_aead_aes256gcm_KEYBYTES = nacl.crypto_aead_aes256gcm_keybytes()
+    crypto_aead_aes256gcm_NPUBBYTES = nacl.crypto_aead_aes256gcm_npubbytes()
+    crypto_aead_aes256gcm_ABYTES = nacl.crypto_aead_aes256gcm_abytes()
+    crypto_aead_chacha20poly1305_ietf_KEYBYTES = nacl.crypto_aead_chacha20poly1305_ietf_keybytes()
+    crypto_aead_chacha20poly1305_ietf_NPUBBYTES = nacl.crypto_aead_chacha20poly1305_ietf_npubbytes()
+    crypto_aead_chacha20poly1305_ietf_ABYTES = nacl.crypto_aead_chacha20poly1305_ietf_abytes()
+    HAS_AEAD = True
+except AttributeError:
+    HAS_AEAD = False
+
 crypto_box_SECRETKEYBYTES = nacl.crypto_box_secretkeybytes()
 crypto_box_SEEDBYTES = nacl.crypto_box_seedbytes()
 crypto_box_PUBLICKEYBYTES = nacl.crypto_box_publickeybytes()
@@ -102,7 +118,6 @@ crypto_box_NONCEBYTES = nacl.crypto_box_noncebytes()
 crypto_box_ZEROBYTES = nacl.crypto_box_zerobytes()
 crypto_box_BOXZEROBYTES = nacl.crypto_box_boxzerobytes()
 crypto_box_BEFORENMBYTES = nacl.crypto_box_beforenmbytes()
-crypto_box_SEALBYTES = nacl.crypto_box_sealbytes()
 crypto_scalarmult_BYTES = nacl.crypto_scalarmult_bytes()
 crypto_scalarmult_SCALARBYTES = nacl.crypto_scalarmult_scalarbytes()
 crypto_sign_BYTES = nacl.crypto_sign_bytes()
@@ -254,6 +269,8 @@ def crypto_box_seal(msg, pk):
 
     enc_msg = nacl.crypto_box_seal('secret message', <public key string>)
     '''
+    if not HAS_SEAL:
+        raise ValueError('Underlying Sodium library does not support sealed boxes')
     if len(pk) != crypto_box_PUBLICKEYBYTES:
         raise ValueError('Invalid public key')
     if not isinstance(msg, bytes):
@@ -269,13 +286,15 @@ def crypto_box_seal_open(ctxt, pk, sk):
     '''
     Decrypts a message given the receiver's public and private key.
     '''
+    if not HAS_SEAL:
+        raise ValueError('Underlying Sodium library does not support sealed boxes')
     if len(pk) != crypto_box_PUBLICKEYBYTES:
         raise ValueError('Invalid public key')
     if len(sk) != crypto_box_SECRETKEYBYTES:
         raise ValueError('Invalid secret key')
     if not isinstance(ctxt, bytes):
         raise TypeError('Message must be bytes')
-    
+
     c = ctypes.create_string_buffer(len(ctxt) - crypto_box_SEALBYTES)
     ret = nacl.crypto_box_seal_open(c, ctxt, ctypes.c_ulonglong(len(ctxt)), pk, sk)
     if ret:
@@ -403,6 +422,135 @@ def crypto_secretbox_open(ctxt, nonce, key):
         raise ValueError('Failed to decrypt message')
     return msg.raw[crypto_secretbox_ZEROBYTES:]
 
+# Authenticated Symmetric Encryption with Additional Data
+
+
+def crypto_aead_aes256gcm_encrypt(message, aad, nonce, key):
+    """Encrypts and authenticates a message with public additional data using the given secret key, and nonce
+
+    Args:
+        message (bytes): a message to encrypt
+        aad  (bytes): additional public data to authenticate
+        nonce (bytes): nonce, does not have to be confidential must be
+            `crypto_aead_aes256gcm_NPUBBYTES` in length
+        key (bytes): secret key, must be `crypto_aead_aes256gcm_KEYBYTES` in
+            length
+
+    Returns:
+        bytes: the ciphertext
+
+    Raises:
+        ValueError: if arguments' length is wrong or the operation has failed.
+    """
+    if len(key) != crypto_aead_aes256gcm_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_aead_aes256gcm_NPUBBYTES:
+        raise ValueError('Invalid nonce')
+
+    length = len(message) + crypto_aead_aes256gcm_ABYTES
+    clen = ctypes.c_ulonglong()
+    c = ctypes.create_string_buffer(length)
+    ret = nacl.crypto_aead_aes256gcm_encrypt(
+        c, ctypes.pointer(clen),
+        message, ctypes.c_ulonglong(len(message)),
+        aad, ctypes.c_ulonglong(len(aad)),
+        None,
+        nonce, key)
+    if ret:
+        raise ValueError('Failed to encrypt message')
+    return c.raw
+
+
+def crypto_aead_chacha20poly1305_ietf_encrypt(message, aad, nonce, key):
+    """Encrypts and authenticates a message with public additional data using the given secret key, and nonce
+
+    Args:
+        message (bytes): a message to encrypt
+        aad  (bytes): additional public data to authenticate
+        nonce (bytes): nonce, does not have to be confidential must be
+            `crypto_aead_chacha20poly1305_ietf_NPUBBYTES` in length
+        key (bytes): secret key, must be `crypto_aead_chacha20poly1305_ietf_KEYBYTES` in
+            length
+
+    Returns:
+        bytes: the ciphertext
+
+    Raises:
+        ValueError: if arguments' length is wrong or the operation has failed.
+    """
+    if len(key) != crypto_aead_chacha20poly1305_ietf_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_aead_chacha20poly1305_ietf_NPUBBYTES:
+        raise ValueError('Invalid nonce')
+
+    length = len(message) + crypto_aead_chacha20poly1305_ietf_ABYTES
+    clen = ctypes.c_ulonglong()
+    c = ctypes.create_string_buffer(length)
+    ret = nacl.crypto_aead_chacha20poly1305_ietf_encrypt(
+        c, ctypes.pointer(clen),
+        message, ctypes.c_ulonglong(len(message)),
+        aad, ctypes.c_ulonglong(len(aad)),
+        None,
+        nonce, key)
+    if ret:
+        raise ValueError('Failed to encrypt message')
+    return c.raw
+
+
+def crypto_aead_aes256gcm_decrypt(ctxt, aad, nonce, key):
+    """
+    Decrypts a ciphertext ctxt given the key, nonce, and aad. If the aad
+    or ciphertext were altered then the decryption will fail.
+    """
+    if len(key) != crypto_aead_aes256gcm_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_aead_aes256gcm_NPUBBYTES:
+        raise ValueError('Invalid nonce')
+
+    length = len(ctxt)-crypto_aead_aes256gcm_ABYTES
+    mlen = ctypes.c_ulonglong()
+    m = ctypes.create_string_buffer(length)
+
+    ret = nacl.crypto_aead_aes256gcm_decrypt(
+        m, mlen,
+        None,
+        ctxt, ctypes.c_ulonglong(len(ctxt)),
+        aad, ctypes.c_ulonglong(len(aad)),
+        nonce, key)
+    if ret:
+        raise ValueError('Failed to decrypt message')
+    return m.raw
+
+
+def crypto_aead_chacha20poly1305_ietf_decrypt(ctxt, aad, nonce, key):
+    """
+    Decrypts a ciphertext ctxt given the key, nonce, and aad. If the aad
+    or ciphertext were altered then the decryption will fail.
+    """
+    if len(key) != crypto_aead_chacha20poly1305_ietf_KEYBYTES:
+        raise ValueError('Invalid key')
+
+    if len(nonce) != crypto_aead_chacha20poly1305_ietf_NPUBBYTES:
+        raise ValueError('Invalid nonce')
+
+    length = len(ctxt)-crypto_aead_chacha20poly1305_ietf_ABYTES
+    mlen = ctypes.c_ulonglong()
+    m = ctypes.create_string_buffer(length)
+
+    ret = nacl.crypto_aead_chacha20poly1305_ietf_decrypt(
+        m, mlen,
+        None,
+        ctxt, ctypes.c_ulonglong(len(ctxt)),
+        aad, ctypes.c_ulonglong(len(aad)),
+        nonce, key)
+    if ret:
+        raise ValueError('Failed to decrypt message')
+    return m.raw
+
+
 # Symmetric Encryption
 
 
